package LightController;

import java.awt.Frame;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.event.MouseListener;
import java.awt.event.MouseEvent;
import java.awt.event.WindowListener;
import java.awt.event.WindowEvent;

import java.util.concurrent.TimeUnit;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.Condition;

import TronAdapter.*;

public class LightController extends TronThread
    implements MouseListener, WindowListener
{
    private enum Loc { Idle, Grasped, Alert, Hold }
    private Loc location;

    static final int epsilon=200;
    static final int delta=500;

    boolean alive = false;

    long startTime;
    long aTime; //time of activation

    TestFixture testFixture = null;
    Dimmer dimmer = null;

    Lock lock = null;
    Condition cond = null;

    final int mutant, levelCount;

    public LightController(int mutant, int levelCount)
    {
	super("LightController");
	assert(levelCount>0);
	location = Loc.Idle;
	this.mutant = mutant;
	this.levelCount = levelCount;
	testFixture = new TestFixture(9999, this);
	dimmer = new Dimmer(aTime, levelCount, mutant);
	addLevelListener(testFixture);
	dimmer.start();
    }

    public void run()
    {
	try { synchronized(dimmer) { while (!dimmer.alive) dimmer.wait(); }}
	catch(InterruptedException e){ System.err.println(e); System.exit(1);}

	lock = new TronLock();
	cond = lock.newCondition();
	lock.lock();

	synchronized (this) { alive = true; notifyAll(); }

	aTime = TronThread.getTimeMillis();
	while (alive) {
	    System.out.println("IFace state: "+location);
	    switch (location) {
	    case Idle:
		try { cond.await(); }
		catch(InterruptedException e) { alive = false; }
		break;
	    case Grasped:
		try { cond.await(epsilon, TimeUnit.MILLISECONDS); }
		catch(InterruptedException e) { alive = false; }
		finally {
		    if (location == Loc.Grasped) location = Loc.Alert;
		}
		break;
	    case Alert:
		try { cond.await(delta-epsilon, TimeUnit.MILLISECONDS); }
		catch (InterruptedException e) { alive = false; }
		finally {
		    if (location == Loc.Alert) {
			location = Loc.Hold;
			dimmer.handleStartHold(startTime+delta);
		    }
		}
		break;
	    case Hold:
		try { cond.await(); }
		catch(InterruptedException e) { alive = false; }
		break;
	    default:
		alive = false;
	    }
	}
	System.out.println("IFace interrupted in "+location);
	dimmer.interrupt();
	testFixture.interrupt();
	lock.unlock();
    }

    public void handleGrasp()
    {
	synchronized (this) {
	    while (!alive) try {wait();}catch(InterruptedException e){}
	}
	lock.lock();
	startTime = getTimeMillis();
	System.out.println("LC: handleGRASP at "+(startTime-aTime));
	switch (location) {
	case Idle:
	    location = Loc.Grasped;
	    cond.signalAll();
	    break;
	default:
	    System.out.println("IFace: cannot accept handleGrasp in "+
			       location);
	    //	    interrupt();
	}
	lock.unlock();
    }

    public void handleRelease()
    {
	synchronized (this) {
	    while (!alive) try {wait();}
	    catch(InterruptedException e){}
	}
	lock.lock();
	System.out.println("LC: handleRELEASE at "+ (getTimeMillis()-aTime));
	switch (location) {
	case Grasped:
	    location = Loc.Idle;
	    cond.signalAll();
	    break;
	case Alert:
	    location = Loc.Idle;
	    cond.signalAll();
	    dimmer.handleTouch();
	    break;
	case Hold:
	    location = Loc.Idle;
	    cond.signalAll();
	    dimmer.handleEndHold();
	    break;
	default:
	    System.out.println("IFace: cannot accept handleRelease in "+
			       location);
	    //	    interrupt();
	}
	lock.unlock();
    }

    public void addLevelListener(LevelListener listener) {
	dimmer.addLevelListener(listener);
    }

    public void mouseClicked(MouseEvent e){}
    public void mouseEntered(MouseEvent e){}
    public void mouseExited(MouseEvent e){}
    public void mousePressed(MouseEvent e)
    {
	if (realtime()) handleGrasp();
	else {
	    System.err.println("##### Simulated time framework does not allow outside interraction. #####");
	    java.awt.Toolkit.getDefaultToolkit().beep();
	}
    }
    public void mouseReleased(MouseEvent e)
    {
	if (realtime()) handleRelease();
	else {
	    System.err.println("##### Simulated time framework does not allow outside interraction. #####");
	    java.awt.Toolkit.getDefaultToolkit().beep();
	}
    }

    public void windowActivated(WindowEvent e) {}
    public void windowClosed(WindowEvent e) {}
    public void windowClosing(WindowEvent e) { System.exit(0); }
    public void windowDeactivated(WindowEvent e) {}
    public void windowDeiconified(WindowEvent e) {}
    public void windowIconified(WindowEvent e) {}
    public void windowOpened(WindowEvent e) {}

    public static void main(String args[])
    {
	int mutant = 0;
	boolean noGUI = false;
	int i = 0;
	while (i<args.length) {
	    if ("-N".equals(args[i])){
		noGUI = true; i++;
	    } else if ("-M".equals(args[i])){
		if (i+1<args.length) mutant = Integer.parseInt(args[i+1]);
		else {
		    System.err.println("Specify mutant id.");
		    return ;
		}
		i += 2 ;
	    } else if ("-C".equals(args[i])) {
		if (i+2<args.length) {
		    int port = Integer.parseInt(args[i+2]);
		    if (port <= 0) {
			System.err.println("The specified port ("+args[i+2]+
					   ") is not in valid range.");
			return;
		    }
		    System.out.println("Setting simulated clock to "+
				       args[i+1]+":"+port);
		    TronThread.setRemoteClock(args[i+1], port);
		    i += 3;
		} else {
		    System.err.println("Specify simulated clock, like: "+
				       "-C localhost 6521");
		    return ;
		}
	    } else {
		System.err.println("Uninterpreted option: "+args[i]);
		i++;
	    }
	}
	if (mutant>=0 && mutant<=2) {
	    LightController lc = new LightController(mutant, 10);
	    System.out.println("LC constructed");
	    if (!noGUI) {
		Frame f = new Frame("LightController");
		f.addWindowListener(lc);
		LevelHistory lh = new LevelHistory(10, 0);
		lh.addMouseListener(lc);
		lc.addLevelListener(lh);
		LevelBar lb = new LevelBar(10, 0);
		lb.addMouseListener(lc);
		lc.addLevelListener(lb);
		LevelLamp lamp = new LevelLamp(10, 0);
		lamp.addMouseListener(lc);
		lc.addLevelListener(lamp);

		GridBagLayout gridbag = new GridBagLayout();
		GridBagConstraints c = new GridBagConstraints();
		f.setLayout(gridbag);
		c.fill = GridBagConstraints.BOTH; c.weighty = 1.0;
		c.weightx = 9.0;
		gridbag.setConstraints(lh, c); f.add(lh);
		c.weightx = 1.0;
		gridbag.setConstraints(lb, c); f.add(lb);
		c.weightx = 5.0;
		c.gridwidth = GridBagConstraints.REMAINDER; //end row
		gridbag.setConstraints(lamp, c); f.add(lamp);
		f.setBounds(0, 0, 400, 202); f.setVisible(true);
		lh.waitForReady();
	    }
	    lc.start();
	    System.out.println("LC started");
	    try { lc.join(); }
	    catch (InterruptedException e) {}
	    System.out.println("LightController terminated");
	} else System.out.println("Unknown Mutant");
    }
}
