package TronAdapter;

import java.net.Socket;
import java.net.UnknownHostException;
import java.net.ConnectException;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;

import java.util.concurrent.TimeUnit;

/**
 * Simulated time framework thread. Use TronThread instead of Thread if you 
 * want your application to be able to run both: realtime and simulated time.
 *<p>
 * Simulated time framework assumptions:
 *<p>
 * All threads are created via TronThread (the initial thread from main() 
 * is useless).
 *<p>
 * The time spent doing computations is negligible and virtually is zero.
 *<p>
 * Time ellapses only during explicit delays. Use TronCondition.await() 
 * methods.
 *<p>
 * Monitor paradigm is supported through TronLock and TronCondition, i.e. lock 
 * TronLock object, compute, potentially call TronCondition.await or 
 * TronCondition.signal()/signalAll().
 *<p>
 */
public class TronThread extends Thread
{
    static String host = null;
    static int port;
    static Object lock = new Object();

    public static boolean realtime() { return (host == null); }
    public static boolean simtime() { return (host != null); }

    public static void setRemoteClock(String _host, int _port) {
	synchronized (lock) {
	    host = _host;
	    port = _port;
	}
    }

    public TronThread(){ super(); }
    public TronThread(String name){ super(name); }
    public TronThread(Runnable r){ super(r); }
    public TronThread(Runnable r, String name){ super(r, name); }

    static final char C_OK = 0;
    static final char C_MutexInit = 1;
    static final char C_MutexDest = 2;
    static final char C_MutexLock = 3;
    static final char C_MutexUnlock = 4;
    static final char C_CondInit = 5;
    static final char C_CondDest = 6;
    static final char C_Wait = 7;
    static final char C_TimedWait = 8;
    static final char C_Delay = 9;// timedwait from now
    static final char C_Signal = 10;
    static final char C_Broadcast = 11;
    static final char C_GetTime = 12;
    static final char C_TimedOut = 13;
    static final char C_Signalled = 14;
    static final char C_Busy = 15;
    static final char C_Other = 16;
    static final char C_Quit = 127;

    private boolean abort = false;
    private DataInputStream is = null;
    private DataOutputStream os = null;

    public void start()
    {
	String myhost = null;
	int myport = 0;
	synchronized (lock) {
	    if (host != null) {
		myhost = host;
		myport = port;
	    }
	}
	if (myhost != null) {
	    int tries = 10;
	    while (tries > 0) try {
		tries--;
		System.out.println("Trying to connect to clock #"+tries);
		Socket socket = new Socket(myhost, myport);
		socket.setTcpNoDelay(true);
		is = new DataInputStream(socket.getInputStream());
		os = new DataOutputStream(socket.getOutputStream());
		String id = getName();
		os.writeByte(id.length());
		os.writeBytes(id);
	    } catch (UnknownHostException e) {
		System.err.println(e);
		System.exit(1);
	    } catch (ConnectException e) {
		System.err.println(e);
		try { Thread.sleep(2000); } // let's retry
		catch(InterruptedException ex){}
	    } catch (IOException e) {
		System.err.println(e);
		System.exit(1);
	    } finally {
		if (is != null) {
		    System.out.println("Connected to clock: "+getName());
		    break;
		}
		if (tries <= 0) System.exit(1);
	    }
	}
	super.start();
    }

    public int mutexInit()
    {
	if (os != null)  try {
	    os.writeByte(C_MutexInit);
	    os.flush();
	    return is.readInt();
	} catch (IOException e) {
	    System.err.println(e);
	    System.exit(1);
	}
	return -1;
    }

    public void mutexDestroy(TronLock m)
    {
	if (os != null)  try {
	    os.writeByte(C_MutexDest);
	    os.writeInt(m.id);
	    os.flush();
	} catch (IOException e) {
	    System.err.println(e);
	    System.exit(1);
	}
    }

    public void mutexLock(TronLock m)
    {
	if (os != null)  try {
	    os.writeByte(C_MutexLock);
	    os.writeInt(m.id);
	    os.flush();
	    int res = is.readInt(); // block until result comes out
	    if (res==0) return;
	    else {
		System.err.println("failed to lock mutex: "+res);
		System.exit(1);
	    }
	} catch (IOException e) {
	    System.err.println(e);
	    System.exit(1);
	}
    }

    public void mutexUnlock(TronLock m)
    {
	if (os != null)  try {
	    os.writeByte(C_MutexUnlock);
	    os.writeInt(m.id);
	    os.flush();
	    int res = is.readInt();
	    if (res==0) return;
	    else {
		System.err.println("failed to unlock mutex: "+res);
		System.exit(1);
	    }
	} catch (IOException e) {
	    System.err.println(e);
	    System.exit(1);
	}
    }

    public int condInit()
    {
	if (os != null)  try {
	    os.writeByte(C_CondInit);
	    os.flush();
	    return is.readInt();
	} catch (IOException e) {
	    System.err.println(e);
	    System.exit(1);
	}
	return -1;
    }

    public void condDestroy(TronCondition c)
    {
	if (os != null)  try {
	    os.writeByte(C_CondDest);
	    os.writeInt(c.id);
	    os.flush();
	} catch (IOException e) {
	    System.err.println(e);
	    System.exit(1);
	}
    }

    public void condWait(TronCondition c) throws InterruptedException
    {
	if (os != null) {
	    try {
		os.writeByte(C_Wait);
		os.writeInt(c.id);
		os.writeInt(c.lock.id);
		os.flush();
		is.readInt();// block on reading the result
	    } catch (IOException e) {
		System.err.println(e);
		System.exit(1);
	    }
	} else synchronized(c) { c.wait(); }
    }

    public boolean condDelay(TronCondition c, long time, TimeUnit unit)
	throws InterruptedException
    {
	if (os != null) {
	    int res, sec, micros;
	    switch (unit) {
	    case MICROSECONDS:
		sec = (int)(time / 1000000);
		micros = (int)(time % 1000000);
		break;
	    case MILLISECONDS:
		sec = (int)(time / 1000);
		micros = (int)(time % 1000 * 1000);
		break;
	    case NANOSECONDS:
		sec = (int)(time / 1000000000);
		micros = (int)(time / 1000 % 1000000);
		break;
	    case SECONDS:
		sec = (int) time;
		micros = 0;
		break;
	    default:
		throw new UnsupportedOperationException("unknown timeunit");
	    }
	    try {
		os.writeByte(C_Delay);
		os.writeInt(c.id);
		os.writeInt(c.lock.id);
		os.writeInt(sec);
		os.writeInt(micros);
		os.flush();
		res = is.readInt(); // block on reading the result
		switch (res) {
		case C_Signalled: return true;
		case C_TimedOut: return false;
		default: throw new InterruptedException("Clock error: "+res);
		}
	    } catch (IOException e) {
		System.err.println(e);
		System.exit(1);
	    }
	}
	throw new UnsupportedOperationException("not connected");
    }

    public boolean condWait(TronCondition c, long millis)
	throws InterruptedException
    {
	if (os != null) {
	    int res;
	    try {
		os.writeByte(C_TimedWait);
		os.writeInt(c.id);
		os.writeInt(c.lock.id);
		os.writeInt((int)(millis / 1000));
		os.writeInt((int)(millis % 1000 * 1000));
		os.flush();
		res = is.readInt(); // block on reading the result
		switch (res) {
		case C_Signalled: return true;
		case C_TimedOut: return false;	       
		}
		System.err.println("SimClock error: "+res);
		System.exit(-1);
	    } catch (IOException e) {
		System.err.println(e);
		System.exit(-1);
	    }
	}
	throw new UnsupportedOperationException("not connected");
    }

    public boolean condDelay(TronCondition c, long millis)
	throws InterruptedException
    {
	if (os != null) {
	    int res;
	    try {		
		os.writeByte(C_Delay);
		os.writeInt(c.id);
		os.writeInt(c.lock.id);
		os.writeInt((int)(millis / 1000));
		os.writeInt((int)(millis % 1000 * 1000));
		os.flush();
		res = is.readInt(); // block on reading the result
		switch (res) {
		case C_Signalled: return true;
		case C_TimedOut: return false;	       
		}
		System.err.println("SimClock error: "+res);
		System.exit(-1);
	    } catch (IOException e) {
		System.err.println(e);
		System.exit(1);
	    }
	}
	throw new UnsupportedOperationException("not connected");
    }


    public void condSignal(TronCondition c) {
	if (os != null) {
	    throw new UnsupportedOperationException("not implemented");
	}
	throw new UnsupportedOperationException("not connected");
    }

    public void condBroadcast(TronCondition c)
    {
	if (os != null) {
	    try {
		os.writeByte(C_Broadcast);
		os.writeInt(c.id);
		os.flush();
	    } catch (IOException e) {
		System.err.println(e);
		System.exit(1);
	    }
	} else
	    throw new UnsupportedOperationException("not connected");
    }

    protected long getTime()
    {
	long millis = 0;
	if (os != null) {
	    try {
		os.writeByte(C_GetTime); os.flush();
		millis = is.readInt() * 1000; // seconds
		millis += is.readInt() / 1000; // micro seconds
		return millis;
	    } catch (IOException e) {
		System.err.println(e);
		System.exit(1);
	    }
	}
	throw new UnsupportedOperationException("not connected");
    }

    public static long getTimeMillis()
    {
	if (host == null) return System.currentTimeMillis();
	else {
	    TronThread orig = (TronThread)Thread.currentThread();
	    return orig.getTime();
	}
    }
}
